import asyncio
import json
from typing import Optional

from aiohttp import web, http_websocket
from nacl.signing import VerifyKey
from nacl.utils import random

from v6d0auth import certs
from v6d0auth.appfactory import *
from v6d0auth.cdb import *

__all__ = ('V6D0AuthAppFactory',)


class V6D0AuthAppFactory(AppFactory):
    def __init__(self, cdb: CDB):
        self.cdb = cdb

    def define_routes(self, routes: web.RouteTableDef):
        print(certs.vkey.encode().hex())
        self.cdb.start()

        @routes.get('/')
        async def home(_request: web.Request):
            return web.Response(body='v6d0auth\n')

        async def ws_approve(ws: web.WebSocketResponse):
            nonce = random(16)
            await ws.send_bytes(nonce)
            hhandle, hnonce = json.loads(certs.verify(await ws.receive_bytes()))
            assert hnonce == nonce.hex()
            if hhandle == 'all':
                print('approving all')
                for request in self.cdb.handle_mapping.values():
                    request.approve()
                print('approved all')
            else:
                approved = self.cdb.approve(bytes.fromhex(hhandle))
                await ws.send_bytes(approved)
            await ws.close()

        @routes.get('/approve')
        async def approve(request: web.Request):
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            await ws_approve(ws)
            return ws

        async def requester_for_request(request: web.Request) -> VerifyKey:
            return VerifyKey(await request.read())

        def role_for_request(request: web.Request) -> Optional[str]:
            return request.headers.get('v6role')

        def pushed_for_role(requester: VerifyKey, role: Optional[str]) -> AbstractRequest:
            if role is None:
                return self.cdb.push_requester(requester)
            else:
                return self.cdb.push_role(Role(requester, role))

        async def pushed_for_request(request: web.Request) -> AbstractRequest:
            return pushed_for_role(await requester_for_request(request), role_for_request(request))

        @routes.post('/push')
        async def push(request: web.Request):
            pushed = await pushed_for_request(request)
            timeout = pushed.timeout
            return web.Response(body=str(timeout))

        def pulled_for_role(requester: VerifyKey, role: Optional[str]) -> Optional[bytes]:
            if role is None:
                return self.cdb.pull_requester(requester)
            else:
                return self.cdb.pull_role(Role(requester, role))

        async def pulled_for_request(request: web.Request) -> Optional[bytes]:
            return pulled_for_role(await requester_for_request(request), role_for_request(request))

        @routes.post('/pull')
        async def pull(request: web.Request):
            try:
                pulled = await pulled_for_request(request)
            except KeyError as e:
                raise web.HTTPNotFound from e
            else:
                return web.Response(body=pulled)

        @routes.post('/has_role')
        async def has_role(request: web.Request):
            role = role_for_request(request)
            if role is None:
                raise web.HTTPBadRequest
            return web.Response(
                body=(b'1' if self.cdb.has_role(Role(await requester_for_request(request), role)) else b'')
            )

        async def ws_remove(ws: web.WebSocketResponse):
            nonce = random(16)
            await ws.send_bytes(nonce)
            [hrequester, role], hnonce = json.loads(certs.verify(await ws.receive_bytes()))
            assert hnonce == nonce.hex()
            self.cdb.remove_role(Role(VerifyKey(bytes.fromhex(hrequester)), role))
            await ws.send_bytes(b'0')
            await ws.close()

        @routes.get('/remove_role')
        async def remove_role(request: web.Request):
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            await ws_remove(ws)
            return ws

        def srq_for_role(requester: VerifyKey, role: Optional[str]) -> AbstractRequest:
            if role is None:
                return self.cdb.requester_mapping[requester]
            else:
                return self.cdb.role_mapping[Role(requester, role)]

        async def srq_for_ws(request: web.Request, ws: web.WebSocketResponse) -> AbstractRequest:
            return srq_for_role(VerifyKey(await ws.receive_bytes()), role_for_request(request))

        async def sqr_fail(ws: web.WebSocketResponse, srq: AbstractRequest) -> None:
            srq.force_repair()
            await ws.close(code=http_websocket.WSCloseCode.TRY_AGAIN_LATER)

        async def srq_success(ws: web.WebSocketResponse, approved: bytes) -> None:
            await ws.send_bytes(approved)
            await ws.close()

        async def srq_process(ws: web.WebSocketResponse, srq: AbstractRequest) -> None:
            try:
                approved = await srq.awaitable()
            except asyncio.CancelledError:
                await sqr_fail(ws, srq)
            else:
                await srq_success(ws, approved)

        async def ws_fail(ws: web.WebSocketResponse) -> None:
            await ws.close(code=http_websocket.WSCloseCode.POLICY_VIOLATION)

        async def ws_process(request: web.Request, ws: web.WebSocketResponse) -> None:
            try:
                srq = await srq_for_ws(request, ws)
            except TypeError:
                await ws_fail(ws)
            else:
                await srq_process(ws, srq)

        @routes.get('/pullws')
        async def pullws(request: web.Request):
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            await ws_process(request, ws)
            return ws
