import asyncio
import functools
import heapq
import time
import weakref
from typing import MutableMapping, Optional, Hashable

from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey
from nacl.utils import random
from ptvp35 import Db, KVJson

from v6d0auth import certs
from v6d0auth.config import myroot

__all__ = ('CDB', 'Role', 'AbstractRequest',)

TIMEOUT = 300


@functools.total_ordering
class AbstractRequest:
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop
        self.timeout = time.time() + TIMEOUT
        self.handle: bytes = random(12)
        self._approved: Optional[bytes] = None
        self.cancelled = False
        self.future: asyncio.Future[bytes] = self._loop.create_future()

    def __lt__(self, other):
        if isinstance(other, AbstractRequest):
            return self.timeout < other.timeout
        else:
            return NotImplemented

    def timed_out(self) -> bool:
        return time.time() > self.timeout

    def _approve(self) -> bytes:
        raise NotImplementedError

    def _validate(self) -> None:
        raise NotImplementedError

    def validate(self) -> None:
        assert self._approved is not None
        self._validate()

    def valid(self) -> bool:
        try:
            self.validate()
            return True
        except (AssertionError, ValueError, BadSignatureError):
            return False

    def approve(self) -> bytes:
        approved = self.approved()
        if approved is not None:
            return approved
        self._approved = self._approve()
        self.future.set_result(self._approved)
        print('validating', self.handle.hex())
        self.validate()
        print('approved', self.handle.hex())
        return self._approved

    def repair(self):
        if self.future.done():
            self.future: asyncio.Future[bytes] = self._loop.create_future()
            print('repaired', self.handle.hex(), self.display())

    def force_repair(self):
        if not self.future.done():
            self.future.cancel()
        self.repair()

    def approved(self) -> Optional[bytes]:
        if self.valid():
            return self._approved
        else:
            self.repair()
            return None

    async def awaitable(self) -> bytes:
        approved = self.approved()
        if approved is None:
            approved = await self.future
        return approved

    def cancel(self):
        if not self.future.done():
            self.future.cancel()
            self.cancelled = True

    def display(self) -> str:
        raise NotImplementedError


class SignatureRequest(AbstractRequest):
    def __init__(self, loop: asyncio.AbstractEventLoop, requester: VerifyKey):
        super().__init__(loop)
        self._requester = requester

    def _approve(self) -> bytes:
        return certs.sign(bytes(self._requester))

    def _validate(self) -> None:
        assert self._approved is not None
        assert certs.verify(self._approved) == bytes(self._requester)

    def display(self) -> str:
        return self._requester.encode().hex()


class Role(Hashable):
    def __init__(self, requester: VerifyKey, role: str):
        self._requester = requester
        self._role = role

    def key(self):
        return self._requester.encode().hex(), self._role

    def __hash__(self):
        return hash(self.key())

    def __eq__(self, other):
        if isinstance(other, Role):
            return self.key() == other.key()
        else:
            return NotImplemented

    def display(self):
        return f'{self._requester.encode().hex()}@{self._role}'


class RoleRequest(AbstractRequest):
    def __init__(self, loop: asyncio.AbstractEventLoop, rdb: Db, role: Role):
        super().__init__(loop)
        self._rdb = rdb
        self._role = role

    def _approve(self) -> bytes:
        self._rdb.set_nowait(self._role.key(), True)
        return b'1'

    def _validate(self) -> None:
        assert self._rdb.get(self._role.key(), False)

    def display(self) -> str:
        return self._role.display()


_rdbfile = myroot / 'roles.db'


class CDB:
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.handle_mapping: MutableMapping[bytes, AbstractRequest] = weakref.WeakValueDictionary()
        self.requester_mapping: MutableMapping[VerifyKey, SignatureRequest] = weakref.WeakValueDictionary()
        self.role_mapping: MutableMapping[Role, RoleRequest] = weakref.WeakValueDictionary()
        self.heap: list[AbstractRequest] = []
        self._loop = loop
        self.rdb = Db(_rdbfile, kvfactory=KVJson())

    def _cleanup(self):
        while self.heap and self.heap[0].timed_out():
            request = heapq.heappop(self.heap)
            request.cancel()
            yield request

    def cleanup(self):
        if self.heap:
            print('cleaning up')
        for request in self._cleanup():
            print('cleaned', request.handle.hex())

    def push_abstract(self, request: AbstractRequest):
        heapq.heappush(self.heap, request)
        self.handle_mapping[request.handle] = request
        print('requested', request.handle.hex(), request.display())

    def push_requester(self, requester: VerifyKey) -> SignatureRequest:
        if requester in self.requester_mapping:
            return self.requester_mapping[requester]
        request = SignatureRequest(self._loop, requester)
        self.requester_mapping[requester] = request
        self.push_abstract(request)
        return request

    def push_role(self, role: Role) -> RoleRequest:
        if role in self.role_mapping:
            return self.role_mapping[role]
        request = RoleRequest(self._loop, self.rdb, role)
        self.role_mapping[role] = request
        self.push_abstract(request)
        return request

    def _approve(self, handle: bytes) -> bytes:
        return self.handle_mapping[handle].approve()

    def approve(self, handle: bytes) -> bytes:
        return self._approve(handle)

    def pull_requester(self, vkey: VerifyKey) -> Optional[bytes]:
        return self.requester_mapping[vkey].approved()

    def pull_role(self, role: Role) -> Optional[bytes]:
        return self.role_mapping[role].approved()

    def has_role(self, role: Role) -> bool:
        return self.rdb.get(role.key(), False)

    def remove_role(self, role: Role):
        return self.rdb.set_nowait(role.key(), False)

    async def job(self):
        while True:
            await asyncio.sleep(TIMEOUT)
            self.cleanup()

    def start(self):
        self._loop.create_task(self.job())
