import aiohttp
from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey

from v6d0auth import certs
from v6d0auth.config import caurl, myroot

__all__ = ('request_signature', 'mycert', 'has_role', 'request_role', 'with_role',)


async def request_signature() -> bytes:
    async with aiohttp.ClientSession() as session:
        async with session.post(f'{caurl}/push', data=certs.vkey.encode()) as response:
            if response.status not in [200, 429]:
                raise RuntimeError(response.status)
        async with session.ws_connect(f'{caurl}/pullws') as ws:
            await ws.send_bytes(certs.vkey.encode())
            try:
                return await ws.receive_bytes()
            except TypeError as e:
                raise RuntimeError("signature request failed") from e


_certfile = myroot / 'cert'


async def mycert() -> bytes:
    try:
        cert = _certfile.read_bytes()
        certs.averify(cert)
    except (FileNotFoundError, BadSignatureError):
        cert = await request_signature()
        _certfile.write_bytes(cert)
    return cert


async def has_role(vkey: VerifyKey, role: str):
    async with aiohttp.ClientSession() as session:
        async with session.post(f'{caurl}/has_role', data=vkey.encode(), headers={'v6role': role}) as response:
            return (await response.read()) == b'1'


async def request_role(role: str) -> bytes:
    async with aiohttp.ClientSession() as session:
        async with session.post(f'{caurl}/push', data=certs.vkey.encode(), headers={'v6role': role}) as response:
            if response.status not in [200, 429]:
                raise RuntimeError(response.status)
        async with session.ws_connect(f'{caurl}/pullws', headers={'v6role': role}) as ws:
            await ws.send_bytes(certs.vkey.encode())
            try:
                return await ws.receive_bytes()
            except TypeError as e:
                raise RuntimeError("role request failed") from e


async def with_role(role: str):
    if not await has_role(certs.vkey, role):
        await request_role(role)
    if not await has_role(certs.vkey, role):
        raise RuntimeError("role request failed")
